;;;;;; CDDL HEADER
;;;; The contents of this file are subject to the terms of the
;;;; Common Development and Distribution License, Version 1.1 only
;;;; (the "License").  You may not use this file except in compliance
;;;; with the License.
;;;;
;;;; A copy of the CDDL is available via the Internet at
;;;; https://spdx.org/licenses/CDDL-1.1.html
;;;; and in the LICENSE file in this distribution.
;;;;
;;;; When distributing Covered Code, include this CDDL HEADER in each
;;;; file and include the contents of the LICENSE file from this
;;;; distribution.
;;;;;; /CDDL HEADER

(in-package #:dateprefixcopy)

(defun config-file-path ()
  "Returns the canonical path to dpc's config file, currently
  residing in %LOCALAPPDATA%."
  (concatenate 'string (uiop:getenv "LOCALAPPDATA") "/dpc.ini"))

(defvar *inifile* nil)  ; will be filled later
(defvar *window-x* 100) ; default startup x
(defvar *window-y* 100) ; default startup y

(defun get-dpc-icon ()
  (create-icon 32 32 1 32
               (make-array 4224 :element-type '(unsigned-byte 8))
               *dpc-icon*))

(defun get-dpc-bgimg ()
  (create-bitmap-resource 256 256 1 32 *dpc-drop-image*))

(defun copy-file (infile outfile)
  "The copy routine."
  ;; If infile == outfile, don't do a thing. Else, copy infile to outfile.
  (unless (string= (file-namestring infile) (file-namestring outfile))
    (with-open-file (instream infile
                              :direction :input
                              :element-type '(unsigned-byte 8)
                              :if-does-not-exist nil)
      (when instream
        ;; Binary copy (this will SKIP existing files!):
        (with-open-file (outstream outfile
                                   :direction :output
                                   :element-type '(unsigned-byte 8)
                                   :if-exists nil)
          (when outstream
            (let ((buffer (make-array 8192 :element-type '(unsigned-byte 8))))
              (loop for bytes-read = (read-sequence buffer instream)
                    while (plusp bytes-read)
                    do (write-sequence buffer outstream :end bytes-read)))))))))

(defun generate-output-filename (infile)
  "We'll need to process the input file name first.
  If it already _has_ a yymmdd_ prefix, we'll need to remove that first."
  (let ((filename (file-namestring infile)))
    (if (and (re:match-re "%d+" (subseq filename 0 6)) (char= (char filename 6) #\_))
      (subseq filename 7)
      filename)))

(defun generate-output-path (infile)
  "Generates the target file path for <infile>, i.e. /inpath/yymmdd_infile."
  (let* ((filename (generate-output-filename infile))
         (filepath (namestring (uiop:pathname-directory-pathname infile)))
         (prefix (format-timestring nil (now) :format '((:short-year) (:month 2) (:day 2)))))
    (concatenate 'string filepath prefix "_" filename)))

(defwndproc main-wndproc (hwnd msg wparam lparam)
            "Window messaging (Win32 API and the main reason why dpc is not portable)."
            (switch msg
                    ((const +wm-create+)
                     ;; When the window is created, we need to
                     ;; 1. add the icon
                     ;; 2. enable a dropzone
                     ;; 3. move the window where it was last time
                     ;; 4. make sure we're always on top
                     (set-class-pointer hwnd :icon (get-dpc-icon))
                     (drag-accept-files hwnd t)
                     (move-window hwnd *window-x* *window-y* 150 150)
                     (set-window-pos hwnd :topmost 0 0 0 0 (ftw:logior-consts +SWP-NOMOVE+ +SWP-NOSIZE+)))

                    ((const +wm-paint+)
                     ;; Draw the background image:
                     (with-paint (hwnd hdc)
                                 (with-compatible-dc (hdc-mem hdc)
                                                     (select-object hdc-mem (get-dpc-bgimg))
                                                     (alpha-blend hdc 20 10 hdc-mem 0 0 
                                                                  :width-dest 100 :height-dest 100
                                                                  :width-source 256 :height-source 256))))

                    ((const +wm-dropfiles+)
                     ;; When files are dropped on the window, we'll need to clone them.
                     (let ((hdrop (make-pointer wparam)))
                       (loop for infile in (drag-query-files hdrop)
                             do
                             (copy-file infile (generate-output-path infile)))))

                    ((const +wm-keydown+)
                     ;; Esc = quit
                     (when (= wparam (virtual-key-code :escape)) (post-quit-message)))

                    ((const +wm-destroy+)
                     (drag-accept-files hwnd nil)

                     ;; Config file adventures ahead!
                     (if (probe-file (config-file-path))
                       ;; If it's there, just use it:
                       (setf *inifile* (read-ini-file (config-file-path)))
                       ;; If it's not, create an empty structure first:
                       (setf *inifile* (make-ini :name "dpc")))

                     ;; Store the window position on quit:
                     (let* ((wrect (get-window-rect hwnd))
                            (x (rect-left wrect))
                            (y (rect-top wrect)))
                       (set-ini-property *inifile* '("window" "x") x)
                       (set-ini-property *inifile* '("window" "y") y)
                       (write-ini-file (config-file-path) *inifile*))

                     (post-quit-message)))
            (default-window-proc hwnd msg wparam lparam))

(defun main ()
  "The main window."
  ;; Read stuff from the config file (if it's there):
  (when (probe-file (config-file-path))
    (setf *inifile* (read-ini-file (config-file-path)))

    ;; Get the previous window position:
    (setf *window-x* (get-ini-property *inifile* '("window" "x")))
    (setf *window-y* (get-ini-property *inifile* '("window" "y"))))

  ;; Open the window:
  (default-message-loop (callback main-wndproc)
                        :class-name "DATE_PREFIX_COPY"
                        :title "DatePrefixCopy"
                        :styles (ftw:logior-consts +ws-border+ +ws-caption+)
                        :icon (get-dpc-icon)
                        :icon-small (get-dpc-icon)
                        :width 150 :height 150))
